@@ -38,15 +38,16 @@
 
  * To undertand the concept of chains, consider for instance a chain
  * consisting of #ChamplainFileCache whose next source is
- * #ChamplainNetworkTileSource whose next source is #ChamplainErrorTileSource.
+ * #ChamplainNetworkTileSource whose next source is an error tile source
+ * created with champlain_map_source_factory_create_error_source ().
  * When champlain_map_source_fill_tile() is called on the first object of the
  * chain, #ChamplainFileCache, the cache checks whether it contains the
  * requested tile in its database. If it does, it returns the tile; otherwise,
  * it calls champlain_map_source_fill_tile() on the next source in the chain
  * (#ChamplainNetworkTileSource). The network tile source loads the tile
  * from the network. When successful, it returns the tile; otherwise it requests
- * the tile from the next source in the chain (#ChamplainErrorTileSource).
- * #ChamplainErrorTileSource always generates an error tile, no matter what
+ * the tile from the next source in the chain (error tile source).
+ * The error tile source always generates an error tile, no matter what
  * its next source is.
  */
 
@@ -56,35 +57,22 @@
 
 G_DEFINE_TYPE (ChamplainMapSource, champlain_map_source, G_TYPE_INITIALLY_UNOWNED);
 
-#define GET_PRIVATE(obj)    (G_TYPE_INSTANCE_GET_PRIVATE((obj), CHAMPLAIN_TYPE_MAP_SOURCE, ChamplainMapSourcePrivate))
-
-enum
-{
-  /* normal signals */
-  RELOAD_TILES,
-  LAST_SIGNAL
-};
+#define GET_PRIVATE(obj) \
+  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), CHAMPLAIN_TYPE_MAP_SOURCE, ChamplainMapSourcePrivate))
 
 enum
 {
   PROP_0,
   PROP_NEXT_SOURCE,
+  PROP_RENDERER,
 };
 
-static guint champlain_map_source_signals[LAST_SIGNAL] = { 0, };
-
 struct _ChamplainMapSourcePrivate
 {
   ChamplainMapSource *next_source;
-
-  gulong sig_handler_id;
+  ChamplainRenderer *renderer;
 };
 
-static void reload_tiles_cb (ChamplainMapSource *orig, ChamplainMapSource *self);
-static void on_set_next_source (ChamplainMapSource *map_source,
-    ChamplainMapSource *old_next_source,
-    ChamplainMapSource *new_next_source);
-
 static void
 champlain_map_source_get_property (GObject *object,
     guint prop_id,
@@ -98,30 +86,43 @@ champlain_map_source_get_property (GObject *object,
     case PROP_NEXT_SOURCE:
       g_value_set_object (value, priv->next_source);
       break;
+
+    case PROP_RENDERER:
+      g_value_set_object (value, priv->renderer);
+      break;
+
     default:
-      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
 }
 
+
 static void
 champlain_map_source_set_property (GObject *object,
     guint prop_id,
     const GValue *value,
     GParamSpec *pspec)
 {
-  ChamplainMapSource *map_source = CHAMPLAIN_MAP_SOURCE(object);
+  ChamplainMapSource *map_source = CHAMPLAIN_MAP_SOURCE (object);
 
   switch (prop_id)
     {
     case PROP_NEXT_SOURCE:
       champlain_map_source_set_next_source (map_source,
-                                            g_value_get_object (value));
+          g_value_get_object (value));
       break;
+
+    case PROP_RENDERER:
+      champlain_map_source_set_renderer (map_source,
+          g_value_get_object (value));
+      break;
+
     default:
-      G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
 }
 
+
 static void
 champlain_map_source_dispose (GObject *object)
 {
@@ -134,26 +135,36 @@ champlain_map_source_dispose (GObject *object)
       priv->next_source = NULL;
     }
 
+  if (priv->renderer)
+    {
+      g_object_unref (priv->renderer);
+
+      priv->renderer = NULL;
+    }
+
   G_OBJECT_CLASS (champlain_map_source_parent_class)->dispose (object);
 }
 
+
 static void
 champlain_map_source_finalize (GObject *object)
 {
   G_OBJECT_CLASS (champlain_map_source_parent_class)->finalize (object);
 }
 
+
 static void
-champlain_map_source_constructed  (GObject *object)
+champlain_map_source_constructed (GObject *object)
 {
   if (G_OBJECT_CLASS (champlain_map_source_parent_class)->constructed)
     G_OBJECT_CLASS (champlain_map_source_parent_class)->constructed (object);
 }
 
+
 static void
 champlain_map_source_class_init (ChamplainMapSourceClass *klass)
 {
-  GObjectClass* object_class = G_OBJECT_CLASS (klass);
+  GObjectClass *object_class = G_OBJECT_CLASS (klass);
   GParamSpec *pspec;
 
   g_type_class_add_private (klass, sizeof (ChamplainMapSourcePrivate));
@@ -174,38 +185,37 @@ champlain_map_source_class_init (ChamplainMapSourceClass *klass)
   klass->get_projection = NULL;
 
   klass->fill_tile = NULL;
-  klass->on_set_next_source = on_set_next_source;
 
   /**
-  * ChamplainMapSource:next-source:
-  *
-  * Next source in the loading chain.
-  *
-  * Since: 0.6
-  */
+   * ChamplainMapSource:next-source:
+   *
+   * Next source in the loading chain.
+   *
+   * Since: 0.6
+   */
   pspec = g_param_spec_object ("next-source",
-                               "Next Source",
-                               "Next source in the loading chain",
-                               CHAMPLAIN_TYPE_MAP_SOURCE,
-                               G_PARAM_READWRITE);
+      "Next Source",
+      "Next source in the loading chain",
+      CHAMPLAIN_TYPE_MAP_SOURCE,
+      G_PARAM_READWRITE);
   g_object_class_install_property (object_class, PROP_NEXT_SOURCE, pspec);
 
   /**
-  * ChamplainMapSource::reload-tiles:
-  * @map_source: the #ChamplainMapSource that received the signal
-  *
-  * The ChamplainMapSource::reload-tiles signal is emitted when the map source
-  * changed its style or data
-  *
-  * Since: 0.6
-  */
-  champlain_map_source_signals[RELOAD_TILES] =
-    g_signal_new ("reload-tiles", G_OBJECT_CLASS_TYPE (object_class),
-                  G_SIGNAL_RUN_LAST, 0, NULL, NULL,
-                  g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
-                  0, NULL);
+   * ChamplainMapSource:renderer:
+   *
+   * Renderer used for tiles rendering.
+   *
+   * Since: 0.8
+   */
+  pspec = g_param_spec_object ("renderer",
+      "Tile renderer",
+      "Tile renderer used to render tiles",
+      CHAMPLAIN_TYPE_RENDERER,
+      G_PARAM_READWRITE);
+  g_object_class_install_property (object_class, PROP_RENDERER, pspec);
 }
 
+
 static void
 champlain_map_source_init (ChamplainMapSource *map_source)
 {
@@ -214,16 +224,17 @@ champlain_map_source_init (ChamplainMapSource *map_source)
   map_source->priv = priv;
 
   priv->next_source = NULL;
-  priv->sig_handler_id = 0;
+  priv->renderer = NULL;
 }
 
+
 /**
  * champlain_map_source_get_next_source:
  * @map_source: a #ChamplainMapSource
  *
  * Get the next source in the chain.
  *
- * Returns: the next source in the chain.
+ * Returns: (transfer none): the next source in the chain.
  *
  * Since: 0.6
  */
@@ -235,31 +246,25 @@ champlain_map_source_get_next_source (ChamplainMapSource *map_source)
   return map_source->priv->next_source;
 }
 
-static
-void reload_tiles_cb (G_GNUC_UNUSED ChamplainMapSource *orig, ChamplainMapSource *self)
+
+/**
+ * champlain_map_source_get_renderer:
+ * @map_source: a #ChamplainMapSource
+ *
+ * Get the renderer used for tiles rendering.
+ *
+ * Returns: (transfer none): the renderer.
+ *
+ * Since: 0.8
+ */
+ChamplainRenderer *
+champlain_map_source_get_renderer (ChamplainMapSource *map_source)
 {
-  /* propagate the signal up the chain */
-  g_signal_emit_by_name (self, "reload-tiles", NULL);
+  g_return_val_if_fail (CHAMPLAIN_IS_MAP_SOURCE (map_source), NULL);
+
+  return map_source->priv->renderer;
 }
 
-static void
-on_set_next_source (ChamplainMapSource *map_source,
-    ChamplainMapSource *old_next_source,
-    ChamplainMapSource *new_next_source)
-{
-  ChamplainMapSourcePrivate *priv = map_source->priv;
-  if (old_next_source)
-    {
-      if (g_signal_handler_is_connected (old_next_source, priv->sig_handler_id))
-        g_signal_handler_disconnect (old_next_source, priv->sig_handler_id);
-    }
-
-  if (new_next_source)
-    {
-      priv->sig_handler_id = g_signal_connect (new_next_source, "reload-tiles",
-                             G_CALLBACK (reload_tiles_cb), map_source);
-    }
-}
 
 /**
  * champlain_map_source_set_next_source:
@@ -278,8 +283,6 @@ champlain_map_source_set_next_source (ChamplainMapSource *map_source,
 
   ChamplainMapSourcePrivate *priv = map_source->priv;
 
-  CHAMPLAIN_MAP_SOURCE_GET_CLASS (map_source)->on_set_next_source (map_source, priv->next_source, next_source);
-
   if (priv->next_source != NULL)
     g_object_unref (priv->next_source);
 
@@ -295,6 +298,35 @@ champlain_map_source_set_next_source (ChamplainMapSource *map_source,
   g_object_notify (G_OBJECT (map_source), "next-source");
 }
 
+
+/**
+ * champlain_map_source_set_renderer:
+ * @map_source: a #ChamplainMapSource
+ * @renderer: the renderer
+ *
+ * Sets the renderer used for tiles rendering.
+ *
+ * Since: 0.8
+ */
+void
+champlain_map_source_set_renderer (ChamplainMapSource *map_source,
+    ChamplainRenderer *renderer)
+{
+  g_return_if_fail (CHAMPLAIN_IS_MAP_SOURCE (map_source));
+  g_return_if_fail (CHAMPLAIN_IS_RENDERER (renderer));
+
+  ChamplainMapSourcePrivate *priv = map_source->priv;
+
+  if (priv->renderer != NULL)
+    g_object_unref (priv->renderer);
+
+  g_object_ref_sink (renderer);
+  priv->renderer = renderer;
+
+  g_object_notify (G_OBJECT (map_source), "renderer");
+}
+
+
 /**
  * champlain_map_source_get_id:
  * @map_source: a #ChamplainMapSource
@@ -313,6 +345,7 @@ champlain_map_source_get_id (ChamplainMapSource *map_source)
   return CHAMPLAIN_MAP_SOURCE_GET_CLASS (map_source)->get_id (map_source);
 }
 
+
 /**
  * champlain_map_source_get_name:
  * @map_source: a #ChamplainMapSource
@@ -331,6 +364,7 @@ champlain_map_source_get_name (ChamplainMapSource *map_source)
   return CHAMPLAIN_MAP_SOURCE_GET_CLASS (map_source)->get_name (map_source);
 }
 
+
 /**
  * champlain_map_source_get_license:
  * @map_source: a #ChamplainMapSource
@@ -349,6 +383,7 @@ champlain_map_source_get_license (ChamplainMapSource *map_source)
   return CHAMPLAIN_MAP_SOURCE_GET_CLASS (map_source)->get_license (map_source);
 }
 
+
 /**
  * champlain_map_source_get_license_uri:
  * @map_source: a #ChamplainMapSource
@@ -367,6 +402,7 @@ champlain_map_source_get_license_uri (ChamplainMapSource *map_source)
   return CHAMPLAIN_MAP_SOURCE_GET_CLASS (map_source)->get_license_uri (map_source);
 }
 
+
 /**
  * champlain_map_source_get_min_zoom_level:
  * @map_source: a #ChamplainMapSource
@@ -385,6 +421,7 @@ champlain_map_source_get_min_zoom_level (ChamplainMapSource *map_source)
   return CHAMPLAIN_MAP_SOURCE_GET_CLASS (map_source)->get_min_zoom_level (map_source);
 }
 
+
 /**
  * champlain_map_source_get_max_zoom_level:
  * @map_source: a #ChamplainMapSource
@@ -403,6 +440,7 @@ champlain_map_source_get_max_zoom_level (ChamplainMapSource *map_source)
   return CHAMPLAIN_MAP_SOURCE_GET_CLASS (map_source)->get_max_zoom_level (map_source);
 }
 
+
 /**
  * champlain_map_source_get_tile_size:
  * @map_source: a #ChamplainMapSource
@@ -421,6 +459,7 @@ champlain_map_source_get_tile_size (ChamplainMapSource *map_source)
   return CHAMPLAIN_MAP_SOURCE_GET_CLASS (map_source)->get_tile_size (map_source);
 }
 
+
 /**
  * champlain_map_source_get_projection:
  * @map_source: a #ChamplainMapSource
@@ -439,6 +478,7 @@ champlain_map_source_get_projection (ChamplainMapSource *map_source)
   return CHAMPLAIN_MAP_SOURCE_GET_CLASS (map_source)->get_projection (map_source);
 }
 
+
 /**
  * champlain_map_source_get_x:
  * @map_source: a #ChamplainMapSource
@@ -459,10 +499,11 @@ champlain_map_source_get_x (ChamplainMapSource *map_source,
 {
   g_return_val_if_fail (CHAMPLAIN_IS_MAP_SOURCE (map_source), 0);
 
-  // FIXME: support other projections
+  /* FIXME: support other projections */
   return ((longitude + 180.0) / 360.0 * pow (2.0, zoom_level)) * champlain_map_source_get_tile_size (map_source);
 }
 
+
 /**
  * champlain_map_source_get_y:
  * @map_source: a #ChamplainMapSource
@@ -483,12 +524,12 @@ champlain_map_source_get_y (ChamplainMapSource *map_source,
 {
   g_return_val_if_fail (CHAMPLAIN_IS_MAP_SOURCE (map_source), 0);
 
-  // FIXME: support other projections
-  return ((1.0 - log (tan (latitude * M_PI / 180.0) + 1.0 /
-                      cos (latitude * M_PI / 180.0)) /
-           M_PI) / 2.0 * pow (2.0, zoom_level)) * champlain_map_source_get_tile_size (map_source);
+  /* FIXME: support other projections */
+  return ((1.0 - log (tan (latitude * M_PI / 180.0) + 1.0 / cos (latitude * M_PI / 180.0)) / M_PI) / 
+          2.0 * pow (2.0, zoom_level)) * champlain_map_source_get_tile_size (map_source);
 }
 
+
 /**
  * champlain_map_source_get_longitude:
  * @map_source: a #ChamplainMapSource
@@ -508,12 +549,12 @@ champlain_map_source_get_longitude (ChamplainMapSource *map_source,
     guint x)
 {
   g_return_val_if_fail (CHAMPLAIN_IS_MAP_SOURCE (map_source), 0);
-  //ChamplainMapSourcePrivate *priv = map_source->priv;
-  // FIXME: support other projections
-  gdouble dx = (float)x / champlain_map_source_get_tile_size (map_source);
+  /* FIXME: support other projections */
+  gdouble dx = (float) x / champlain_map_source_get_tile_size (map_source);
   return dx / pow (2.0, zoom_level) * 360.0 - 180;
 }
 
+
 /**
  * champlain_map_source_get_latitude:
  * @map_source: a #ChamplainMapSource
@@ -533,13 +574,13 @@ champlain_map_source_get_latitude (ChamplainMapSource *map_source,
     guint y)
 {
   g_return_val_if_fail (CHAMPLAIN_IS_MAP_SOURCE (map_source), 0);
-  //ChamplainMapSourcePrivate *priv = map_source->priv;
-  // FIXME: support other projections
-  gdouble dy = (float)y / champlain_map_source_get_tile_size (map_source);
+  /* FIXME: support other projections */
+  gdouble dy = (float) y / champlain_map_source_get_tile_size (map_source);
   gdouble n = M_PI - 2.0 * M_PI * dy / pow (2.0, zoom_level);
   return 180.0 / M_PI * atan (0.5 * (exp (n) - exp (-n)));
 }
 
+
 /**
  * champlain_map_source_get_row_count:
  * @map_source: a #ChamplainMapSource
@@ -556,12 +597,11 @@ champlain_map_source_get_row_count (ChamplainMapSource *map_source,
     guint zoom_level)
 {
   g_return_val_if_fail (CHAMPLAIN_IS_MAP_SOURCE (map_source), 0);
-
-  //ChamplainMapSourcePrivate *priv = map_source->priv;
-  // FIXME: support other projections
+  /* FIXME: support other projections */
   return (zoom_level != 0) ? 2 << (zoom_level - 1) : 1;
 }
 
+
 /**
  * champlain_map_source_get_column_count:
  * @map_source: a #ChamplainMapSource
@@ -579,12 +619,11 @@ champlain_map_source_get_column_count (ChamplainMapSource *map_source,
     guint zoom_level)
 {
   g_return_val_if_fail (CHAMPLAIN_IS_MAP_SOURCE (map_source), 0);
-
-  //ChamplainMapSourcePrivate *priv = map_source->priv;
-  // FIXME: support other projections
+  /* FIXME: support other projections */
   return (zoom_level != 0) ? 2 << (zoom_level - 1) : 1;
 }
 
+
 #define EARTH_RADIUS 6378137.0 /* meters, Equatorial radius */
 
 /**
@@ -608,18 +647,20 @@ champlain_map_source_get_meters_per_pixel (ChamplainMapSource *map_source,
 {
   g_return_val_if_fail (CHAMPLAIN_IS_MAP_SOURCE (map_source), 0.0);
 
-  /* Width is in pixels. (1 px)
-     m/px = radius_at_latitude / width_in_pixels
-     k = radius of earth = 6 378.1 km
-     radius_at_latitude = 2π * k * sin (π/2-θ)
-  */
+  /* 
+   * Width is in pixels. (1 px)
+   * m/px = radius_at_latitude / width_in_pixels
+   * k = radius of earth = 6 378.1 km
+   * radius_at_latitude = 2pi * k * sin (pi/2-theta)
+   */
 
   gdouble tile_size = champlain_map_source_get_tile_size (map_source);
-  // FIXME: support other projections
-  return 2.0 * M_PI * EARTH_RADIUS * sin (M_PI/2.0 - M_PI / 180.0 * latitude) /
+  /* FIXME: support other projections */
+  return 2.0 * M_PI *EARTH_RADIUS * sin (M_PI / 2.0 - M_PI / 180.0 * latitude) /
          (tile_size * champlain_map_source_get_row_count (map_source, zoom_level));
 }
 
+
 /**
  * champlain_map_source_fill_tile:
  * @map_source: a #ChamplainMapSource
