@@ -19,11 +19,11 @@
 
 /**
  * SECTION:champlain-network-tile-source
- * @short_description: A map source that downloads tiles from a web server
+ * @short_description: A map source that downloads tile data from a web server
  *
- * This object is specialized for map tiles that can be downloaded
+ * This class is specialized for map tiles that can be downloaded
  * from a web server.  This includes all web based map services such as
- * OpenStreetMap, Google Maps, Yahoo Maps and more.  This object contains
+ * OpenStreetMap, Google Maps, Yahoo Maps and more.  This class contains
  * all mechanisms necessary to download tiles.
  *
  * Some preconfigured network map sources are built-in this library,
@@ -81,26 +81,32 @@ struct _ChamplainNetworkTileSourcePrivate
   SoupSession *soup_session;
 };
 
+typedef struct
+{
+  ChamplainMapSource *map_source;
+  SoupMessage *msg;
+} TileCancelledData;
+
 typedef struct
 {
   ChamplainMapSource *map_source;
   ChamplainTile *tile;
-} TileLoadedCallbackData;
+  TileCancelledData *cancelled_data;
+} TileLoadedData;
 
 typedef struct
 {
   ChamplainMapSource *map_source;
   gchar *etag;
-} TileRenderedCallbackData;
+} TileRenderedData;
 
-typedef struct
-{
-  ChamplainMapSource *map_source;
-  SoupMessage *msg;
-} TileDestroyedCbData;
 
 static void fill_tile (ChamplainMapSource *map_source,
     ChamplainTile *tile);
+static void
+tile_state_notify (ChamplainTile *tile,
+    G_GNUC_UNUSED GParamSpec *pspec,
+    TileCancelledData *data);
 
 static gchar *get_tile_uri (ChamplainNetworkTileSource *source,
     gint x,
@@ -171,6 +177,7 @@ champlain_network_tile_source_dispose (GObject *object)
   if (priv->soup_session)
     {
       soup_session_abort (priv->soup_session);
+      g_object_unref (priv->soup_session);
       priv->soup_session = NULL;
     }
 
@@ -281,7 +288,7 @@ champlain_network_tile_source_init (ChamplainNetworkTileSource *tile_source)
 }
 
 
-/*
+/**
  * champlain_network_tile_source_new_full:
  * @id: the map source's id
  * @name: the map source's name
@@ -292,10 +299,11 @@ champlain_network_tile_source_init (ChamplainNetworkTileSource *tile_source)
  * @tile_size: the map source's tile size (in pixels)
  * @projection: the map source's projection
  * @uri_format: the URI to fetch the tiles from, see #champlain_network_tile_source_set_uri_format
+ * @renderer: the #ChamplainRenderer used to render tiles
  *
  * Constructor of #ChamplainNetworkTileSource.
  *
- * Returns: a constructed #ChamplainNetworkTileSource
+ * Returns: a constructed #ChamplainNetworkTileSource object
  *
  * Since: 0.4
  */
@@ -524,41 +532,24 @@ get_tile_uri (ChamplainNetworkTileSource *tile_source,
 
 
 static void
-tile_destroyed_cb (G_GNUC_UNUSED ChamplainTile *tile,
-    TileDestroyedCbData *data)
-{
-  if (data->map_source && data->msg)
-    {
-      DEBUG ("Canceling tile download");
-      ChamplainNetworkTileSourcePrivate *priv = CHAMPLAIN_NETWORK_TILE_SOURCE (data->map_source)->priv;
-
-      soup_session_cancel_message (priv->soup_session, data->msg, SOUP_STATUS_CANCELLED);
-    }
-}
-
-
-static void
-destroy_cb_data (TileDestroyedCbData *data,
-    G_GNUC_UNUSED GClosure *closure)
-{
-  if (data->map_source)
-    g_object_remove_weak_pointer (G_OBJECT (data->map_source), (gpointer *) &data->map_source);
-
-  g_free (data);
-}
-
-
-static void
-tile_rendered_cb (ChamplainTile *tile, ChamplainRenderCallbackData *data, TileRenderedCallbackData *user_data)
+tile_rendered_cb (ChamplainTile *tile,
+    ChamplainRenderCallbackData *data,
+    TileRenderedData *user_data)
 {
   ChamplainMapSource *map_source = user_data->map_source;
-  ChamplainTileSource *tile_source = CHAMPLAIN_TILE_SOURCE (map_source);
-  ChamplainTileCache *tile_cache = champlain_tile_source_get_cache (tile_source);
-  ChamplainMapSource *next_source = champlain_map_source_get_next_source (map_source);
+  ChamplainMapSource *next_source;
   gchar *etag = user_data->etag;
 
+  g_signal_handlers_disconnect_by_func (tile, tile_rendered_cb, user_data);
+  g_slice_free (TileRenderedData, user_data);
+
+  next_source = champlain_map_source_get_next_source (map_source);
+
   if (!data->error)
     {
+      ChamplainTileSource *tile_source = CHAMPLAIN_TILE_SOURCE (map_source);
+      ChamplainTileCache *tile_cache = champlain_tile_source_get_cache (tile_source);
+
       if (etag != NULL)
         champlain_tile_set_etag (tile, etag);
 
@@ -569,15 +560,12 @@ tile_rendered_cb (ChamplainTile *tile, ChamplainRenderCallbackData *data, TileRe
       champlain_tile_set_state (tile, CHAMPLAIN_STATE_DONE);
       champlain_tile_display_content (tile);
     }
-  else
-    {
-      if (next_source)
-        champlain_map_source_fill_tile (next_source, tile);
-    }
+  else if (next_source)
+    champlain_map_source_fill_tile (next_source, tile);
 
+  g_free (etag);
   g_object_unref (map_source);
-  g_free (user_data);
-  g_signal_handlers_disconnect_by_func (tile, tile_rendered_cb, user_data);
+  g_object_unref (tile);
 }
 
 
@@ -586,34 +574,26 @@ tile_loaded_cb (G_GNUC_UNUSED SoupSession *session,
     SoupMessage *msg,
     gpointer user_data)
 {
-  TileLoadedCallbackData *callback_data = (TileLoadedCallbackData *) user_data;
+  TileLoadedData *callback_data = (TileLoadedData *) user_data;
   ChamplainMapSource *map_source = callback_data->map_source;
   ChamplainTileSource *tile_source = CHAMPLAIN_TILE_SOURCE (map_source);
   ChamplainTileCache *tile_cache = champlain_tile_source_get_cache (tile_source);
   ChamplainMapSource *next_source = champlain_map_source_get_next_source (map_source);
   ChamplainTile *tile = callback_data->tile;
   const gchar *etag;
-  TileRenderedCallbackData *data;
+  TileRenderedData *data;
   ChamplainRenderer *renderer;
 
-
-  if (tile)
-    g_object_remove_weak_pointer (G_OBJECT (tile), (gpointer *) &callback_data->tile);
-
-  g_free (user_data);
+  g_signal_handlers_disconnect_by_func (tile, tile_state_notify, callback_data->cancelled_data);
+  g_slice_free (TileLoadedData, callback_data);
 
   DEBUG ("Got reply %d", msg->status_code);
 
-  if (!tile || msg->status_code == SOUP_STATUS_CANCELLED)
+  if (msg->status_code == SOUP_STATUS_CANCELLED)
     {
-      if (!tile)
-        DEBUG ("Tile destroyed while loading");
-      else
-        DEBUG ("Download of tile %d, %d got cancelled",
-            champlain_tile_get_x (tile), champlain_tile_get_y (tile));
-
-      g_object_unref (map_source);
-      return;
+      DEBUG ("Download of tile %d, %d got cancelled",
+          champlain_tile_get_x (tile), champlain_tile_get_y (tile));
+      goto cleanup;
     }
 
   if (msg->status_code == SOUP_STATUS_NOT_MODIFIED)
@@ -637,12 +617,13 @@ tile_loaded_cb (G_GNUC_UNUSED SoupSession *session,
   etag = soup_message_headers_get (msg->response_headers, "ETag");
   DEBUG ("Received ETag %s", etag);
 
-  data = g_new (TileRenderedCallbackData, 1);
+  renderer = champlain_map_source_get_renderer (map_source);
+  g_return_if_fail (CHAMPLAIN_IS_RENDERER (renderer));
+
+  data = g_slice_new (TileRenderedData);
   data->map_source = map_source;
   data->etag = g_strdup (etag);
 
-  renderer = champlain_map_source_get_renderer (map_source);
-
   g_signal_connect (tile, "render-complete", G_CALLBACK (tile_rendered_cb), data);
 
   champlain_renderer_set_data (renderer, msg->response_body->data, msg->response_body->length);
@@ -653,17 +634,48 @@ tile_loaded_cb (G_GNUC_UNUSED SoupSession *session,
 load_next:
   if (next_source)
     champlain_map_source_fill_tile (next_source, tile);
-  g_object_unref (map_source);
-  return;
+
+  goto cleanup;
 
 finish:
   champlain_tile_set_fade_in (tile, TRUE);
   champlain_tile_set_state (tile, CHAMPLAIN_STATE_DONE);
   champlain_tile_display_content (tile);
+
+cleanup:
+  g_object_unref (tile);
   g_object_unref (map_source);
 }
 
 
+static void
+destroy_cancelled_data (TileCancelledData *data,
+    G_GNUC_UNUSED GClosure *closure)
+{
+  if (data->map_source)
+    g_object_remove_weak_pointer (G_OBJECT (data->map_source), (gpointer *) &data->map_source);
+
+  if (data->msg)
+    g_object_remove_weak_pointer (G_OBJECT (data->msg), (gpointer *) &data->msg);
+
+  g_slice_free (TileCancelledData, data);
+}
+
+
+static void
+tile_state_notify (ChamplainTile *tile,
+    G_GNUC_UNUSED GParamSpec *pspec,
+    TileCancelledData *data)
+{
+  if (champlain_tile_get_state (tile) == CHAMPLAIN_STATE_DONE && data->map_source && data->msg)
+    {
+      DEBUG ("Canceling tile download");
+      ChamplainNetworkTileSourcePrivate *priv = CHAMPLAIN_NETWORK_TILE_SOURCE (data->map_source)->priv;
+      soup_session_cancel_message (priv->soup_session, data->msg, SOUP_STATUS_CANCELLED);
+    }
+}
+
+
 static gchar *
 get_modified_time_string (ChamplainTile *tile)
 {
@@ -699,9 +711,12 @@ fill_tile (ChamplainMapSource *map_source,
   ChamplainNetworkTileSource *tile_source = CHAMPLAIN_NETWORK_TILE_SOURCE (map_source);
   ChamplainNetworkTileSourcePrivate *priv = tile_source->priv;
 
+  if (champlain_tile_get_state (tile) == CHAMPLAIN_STATE_DONE)
+    return;
+
   if (!priv->offline)
     {
-      TileLoadedCallbackData *callback_data;
+      TileLoadedData *callback_data;
       SoupMessage *msg;
       gchar *uri;
 
@@ -709,8 +724,8 @@ fill_tile (ChamplainMapSource *map_source,
           champlain_tile_get_x (tile),
           champlain_tile_get_y (tile),
           champlain_tile_get_zoom_level (tile));
-
       msg = soup_message_new (SOUP_METHOD_GET, uri);
+      g_free (uri);
 
       if (champlain_tile_get_state (tile) == CHAMPLAIN_STATE_LOADED)
         {
@@ -739,22 +754,23 @@ fill_tile (ChamplainMapSource *map_source,
           g_free (date);
         }
 
-      TileDestroyedCbData *tile_destroyed_cb_data = g_new (TileDestroyedCbData, 1);
-      tile_destroyed_cb_data->map_source = map_source;
-      tile_destroyed_cb_data->msg = msg;
+      TileCancelledData *tile_cancelled_data = g_slice_new (TileCancelledData);
+      tile_cancelled_data->map_source = map_source;
+      tile_cancelled_data->msg = msg;
 
-      g_object_add_weak_pointer (G_OBJECT (msg), (gpointer *) &tile_destroyed_cb_data->msg);
-      g_object_add_weak_pointer (G_OBJECT (map_source), (gpointer *) &tile_destroyed_cb_data->map_source);
+      g_object_add_weak_pointer (G_OBJECT (msg), (gpointer *) &tile_cancelled_data->msg);
+      g_object_add_weak_pointer (G_OBJECT (map_source), (gpointer *) &tile_cancelled_data->map_source);
 
-      g_signal_connect_data (tile, "destroy", G_CALLBACK (tile_destroyed_cb),
-          tile_destroyed_cb_data, (GClosureNotify) destroy_cb_data, 0);
+      g_signal_connect_data (tile, "notify::state", G_CALLBACK (tile_state_notify),
+          tile_cancelled_data, (GClosureNotify) destroy_cancelled_data, 0);
 
-      callback_data = g_new (TileLoadedCallbackData, 1);
+      callback_data = g_slice_new (TileLoadedData);
       callback_data->tile = tile;
       callback_data->map_source = map_source;
+      callback_data->cancelled_data = tile_cancelled_data;
 
-      g_object_add_weak_pointer (G_OBJECT (tile), (gpointer *) &callback_data->tile);
       g_object_ref (map_source);
+      g_object_ref (tile);
 
       soup_session_queue_message (priv->soup_session, msg,
           tile_loaded_cb,
